/**
 * KV-32 Torus Field Analyser
 */
//http://s000.tinyupload.com/?file_id=19337604113201776333

//http://s000.tinyupload.com/?del_id=01977271313948736187



"use strict";
this.name        = "torus-field-analyser.raeddi.alaric";
this.author      = "Alaric";
this.copyright   = "2016 Alaric";
this.description = "Raeddi Systems Ltd. Torus Field Analyser";
this.licence     = "CC BY-NC-SA 3.0 AU";
this.version	 = "1.0";



/*
	bogey
	{
		signature:
		entity:
		identity:
		discrepancy:
		expiry		- game clock time when bogey will expire
	}
*/

this.$bogies = [];

this.$private =
{
	lastObservationTime: 0
};

this.$const = 
{
	maximumDiscrepancy:	0.10	// maximum allowable signature-based ID
};

this.$friend =
{
	monitor: "torus-field-monitor.raeddi.alaric",
	events: "torus-field-events.raeddi.alaric"
};

// keep track of player target for ID update

this.$shipSignatureLibrary = [];
this.$torusEverEngaged = false;


this._oneDecimalPlace = function(x)
{
	//var y = Math.floor(x)
	return Math.floor(x) + "." + Math.floor(x * 10 + 0.5) % 10;
}


this._onPlayerTargetChanged = function(oldTarget, newTarget)
{
	if (newTarget) this._recordSignatureForEntity(newTarget);
}


this._getTagForEntity = function(entity)
{
	var propertyName = "$" + this.name;

	if (!entity.isValid) return null;

	if (!entity.hasOwnProperty(propertyName))
	{
		/* create default object */
		entity[propertyName] =
		{
			lastObservationTime:		clock.adjustedSeconds,
			minimumObservedSignature:	Infinity,
			maximumObservedSignature:	0
		};
	}

	return entity[propertyName];
};



/**
 * Displays the first 5 entries of the bogies array on the MFD
 */
this._updateBogiesMFD = function()
{
	if (this.$friend.monitor.$protected.disturbances[0]) 
	{
		this._identifyBogies(this.$friend.monitor.$protected.disturbances[0]);
	}

	var T = clock.adjustedSeconds;
	var Lm = "(" + this._oneDecimalPlace(player.ship.speed / 1000) + "Lm)";
	Lm = "(" + this._oneDecimalPlace(player.ship.speed / player.ship.maxSpeed / 32 * 100) + "%)";

	for (var n = 0; n < this.$mfdTabularText.length; ++n)
	{
		if (!this.$mfdTabularText[n].$expiry) continue;
		if (T >= this.$mfdTabularText[n].$expiry) this.$mfdTabularText[n] = [];
	}

	if (player.alertMassLocked)
	{
		this.$mfdTabularText[8] = [];
		this.$mfdTabularText[9] = [ { width: 14, alignment: "CENTER", text: "Torus drive mass locked" } ];
	}
	
	else if (player.ship.torusEngaged && player.ship.speed < player.ship.maxSpeed * 32)
	{
		this.$mfdTabularText[8] = [];
		this.$mfdTabularText[9] = [ { width: 14, alignment: "CENTER", text: "Torus drive engaging... " + Lm } ];
	}
	else if (player.ship.torusEngaged)
	{
		
		this.$mfdTabularText[8] = [];
		this.$mfdTabularText[9] = [ { width: 14, alignment: "CENTER", text: "Torus drive enaged " + Lm } ];
		

		if (this.$friend.monitor.$public.timeToMassLock !== Infinity) 
			this.$mfdTabularText[9] = [ { width: 14, alignment: "CENTER", text: "Field collapse in " + this._oneDecimalPlace(this.$friend.monitor.$public.timeToMassLock) + " seconds" } ];
		else

		this.$mfdTabularText[9] = (this.$monitor.$public.fieldSynchronised)
			? [ { width: 14, alignment: "CENTER", text: "Torus field stable" } ]
			: [ { width: 14, alignment: "CENTER", text: "Aquiring field alignment" } ]
			;
		

		
	}
	else
	{
		if (!this.$mfdTabularText[8].$expiry) this.$mfdTabularText[8] = [];
		this.$mfdTabularText[9] = [ { width: 14, alignment: "CENTER", text: "Torus drive not enaged" } ];
	}

	// Remove all bogies older than 2 seconds past expiry time
	this.$bogies = this.$bogies.filter(
		function(element, index, array)
		{
			return (T < element.expiry + 2);
		},
		this
	);


	if (this.$bogies.length === 0)
	{
		// erase all MFD lines apart from header and footer
		for (var i = 1; i < 9; ++i) this.$mfdTabularText[i] = [];
		this.$mfdTabularText[4] = [ { width: 14, alignment: "CENTER", text: "KV-32 Ident System Ready" } ];
		this.$mfdTabularText[5] = [ { width: 14, alignment: "CENTER", text: "No anomalies detected" } ];
		return;
	}

	this.$mfdTabularText[1] = [];	
	
	this.$mfdTabularText[2] =
	[	/* column headers */
		{ width: 1, alignment: "RIGHT", text: "#" },   // spacer
		{ width: 0.5, text: "" },   // spacer
		{ width: 6.5, alignment: "LEFT", text: "Bogey" },
		{ width: 6,   alignment: "RIGHT", text: "Discrepancy" }
	];

	//this.$mfdTabularText[8] = [];

	var i;
	var T = clock.adjustedSeconds;
	var blinkOff = false;(T - Math.floor(T) > 0.5);

	//log(this.name, this.$bogies.length);
	for (i = 0; i < this.$bogies.length && i < 5; ++i)
	{
		
		if (T >= this.$bogies[i].expiry)
		{
			this.$mfdTabularText[3 + i] = [ { width: 14, alignment: "CENTER", text: "-- residule decayed -- " } ];
			continue;
		}

		// discrepancy will be null for stars, planets and unidentified
		// disturbances. For anything else, round to 1 decimal place.
		var discrepancy = this.$bogies[i].discrepancy
			? this._oneDecimalPlace(this.$bogies[i].discrepancy * 100) + "%"
			: null
			;

		var blinkOff = (T >= this.$bogies[i].expiry - 3)
			? (T - Math.floor(T) > 0.5)
			: 0
			;

		this.$mfdTabularText[3 + i] = 
		[
			{ width:   1, alignment: "RIGHT", text: i + 1 },
			{ width: 0.5, text: "" },
			{ width:  10, alignment: "LEFT", text: (blinkOff) ? "" : this.$bogies[i].identity },
			
			(discrepancy) 
			? { width: 2.5, alignment: "RIGHT", text: (blinkOff) ? "" : discrepancy}
			: { width: 2.5 - defaultFont.measureString("%"), alignment: "RIGHT", text: (blinkOff) ? "" : "--"}
		];
	}

	/* blank unused rows */
	while (i < 5) this.$mfdTabularText[3 + i++] = [];
}


this._positiveBogeyID = function(entity)
{
	if (!entity.name) return;

	for (var i = 0; i < this.$bogies.length; ++i)
	{
		if (this.$bogies[i].entity === entity)
		{
			this.$bogies[i].identity = entity.name;
		}
	}

	this._updateBogiesMFD();
	var mfdText = worldScripts["alaric-oxp-utilities"]._buildTabularText(this.$mfdTabularText);
	player.ship.setMultiFunctionText("torus-field-analyser", mfdText, false);

}

/**
 * Adds the entity signature to $shipSignatureLibrary if it isn't already there.
 */
this._recordSignatureForEntity = function(entity)
{
	var T = clock.adjustedSeconds;
	if (
		!this.$bogies.some(
			function(element, index, array)
			{ 
				return (element.entity == entity && T < element.expiry)
			}
		)
	) return;

	this._positiveBogeyID(entity);

	var signature = this.$friend.monitor._getSignatureForEntity(entity);
	var bounds = [ 0, this.$shipSignatureLibrary.length ];

	while (bounds[1] > bounds[0]) 
	{
		var pivot = (bounds[0] + bounds[1]) >> 1;
				
		if (signature === this.$shipSignatureLibrary[pivot].signature) break;

		var branch = (signature < this.$shipSignatureLibrary[pivot].signature)
			? 1
			: 0
			;

		bounds[branch] = pivot + 1 - branch;
	}

	// if the two bounds match, this signature did not exist and both values
	// represent the insertion point, so insert the signature now with an
	// empty identities array
	if (bounds[0] === bounds[1])
	{
		this.$shipSignatureLibrary.splice(
			bounds[0],
			0,
			{
				signature: signature,
				identities:	[]
			}
			);
	}

	// the signature now exists
	var match = (bounds[0] + bounds[1]) >> 1;
	var identities = this.$shipSignatureLibrary[match].identities;

	var name = this._getNameForEntity(entity);
	if (identities.indexOf(name) !== -1) return;
	identities.push(name);
		
	this.$mfdTabularText[8] = [ { width: 14, alignment: "CENTER", text: entity.name + " signature recorded" } ];
	this.$mfdTabularText[8].$expiry = clock.adjustedSeconds + 5;

	var mfdText = worldScripts["alaric-oxp-utilities"]._buildTabularText(this.$mfdTabularText);	
	player.ship.setMultiFunctionText("torus-field-analyser", mfdText, false);
	//log(this.name, "this.$shipSignatureLibrary = " + JSON.stringify(this.$shipSignatureLibrary));
}


this._resetMFD = function()
{
	this.$mfdTabularText = 
	[
		[	/* title */
			{ width: 14, alignment: "CENTER", text: "KV-32 Torus Field Analyser" }
		],

		[],	/* spacer */
			
					
		[],	/* bogey #1 */
		[
			{ width: 14, alignment: "CENTER", text: "All hardware and software" }
		],
		[	/* bogey #3 */
			{ width: 14, alignment: "CENTER", text: "\u00a9 3147 Raeddi Systems Ltd." }
		],
		[], /* bogey #4 */

		[],
		[
			{ width: 14, alignment: "CENTER", text: "Are you Raeddi?\u2122" }
		]
		, /* bogey #5 */

		
			
		[],	/* spacer */

		[
			{ width: 14, alignment: "CENTER", text: "Torus drive not engaged" }
		]
	];
}

this.$mfdTabularText = 
[
	[	/* title */
		{ width: 14, alignment: "CENTER", text: "KV-32 Torus Field Analyser" }
	],

	[],	/* spacer */
			
	[	/* column headers */
		{ width: 1, alignment: "RIGHT", text: "#" },   // spacer
		{ width: 0.5, text: "" },   // spacer
		{ width: 6.5, alignment: "LEFT", text: "Bogey" },
		{ width: 6,   alignment: "RIGHT", text: "Discrepancy" }
	],
			
	[],	/* bogey #1 */
	[], /* bogey #2 */
	[	/* bogey #3 */
		{ width: 14, alignment: "CENTER", text: "No contacts recorded" }
	],
	[], /* bogey #4 */
	[], /* bogey #5 */
			
	[],	/* spacer */

	[
		{ width: 14, alignment: "CENTER", text: "Torus drive not engaged" }
	]
];



this._getSaltForEntity = function(entity)
{
	
	var q1 = player.ship.orientation;
	var q2 = entity.orientation;

	var c = q1.w * q2.w + q1.x * q2.x + q1.y * q2.y + q1.z * q2.z;
	return c * Math.sqrt(1 - c * c) * 2;

	
	var up = entity.orientation.vectorUp();
	var right = entity.orientation.vectorRight();
	
	//var v = player.ship.position.directionTo(entity.position);
	var v = entity.position.subtract(player.ship.position).direction();
	


	return v.dot(up) * v.dot(right);

	
}

/**
 * Returns the "observed" signature of the entity. This is the entities
 * actual signature, scaled by some factor based upon ship orientations
 * and distances.
 */
this._getObservedSignature = function(entity)
{
	var tempered = this._getSaltForEntity(entity);	// -1 .. +1
	
	var Dsq = player.ship.position.squaredDistanceTo(entity.position);

	var factor = 1 - Math.pow(player.ship.scannerRange, 2) * 0.9 / Dsq;

	tempered *= factor;

	var signature = this.$friend.monitor._getSignatureForEntity(entity);// + tempered;
	signature += signature * this.$const.maximumDiscrepancy * tempered;
	
	return signature;	
}


this._logSignatureLibrary = function()
{
	var i;

	for (i = 0; i < this.$shipSignatureLibrary.length; ++i)
	{
		var padding = 20 - this.$shipSignatureLibrary[i].$signature.toString().length;
		var j;

		for (j = 0; j < this.$shipSignatureLibrary[i].length; ++j)
		{
			log (
				this.name, 
				this.$shipSignatureLibrary[i][j] +
				new Array(40 - this.$shipSignatureLibrary[i][j].length + padding).join(" ") +
				this.$shipSignatureLibrary[i].$signature + " " +
				(Math.pow(this.$shipSignatureLibrary[i].$signature, 0.5 / Math.PI) * 0.5)
				);
		}

	}
}

this._getNameForEntity = function(entity)
{
	if (entity.shipClassName) return entity.shipClassName;
	else if (entity.name) return entity.name;
	else if (entity.shipUniqueName) return entity.shipUniqueName;
	return null;
}

this._identifyEntity = function(entity)
{
	
	var bogey = 
	{ 
		entity:			entity,
		signature:		null,
		identity:		"Unidentified disturbance",
		discrepancy:	null,
		expiry:			clock.adjustedSeconds + 2
	};

	// Get observed signature and tag object
	var signature = this._getObservedSignature(entity);
	var entityTag = this._getTagForEntity(entity);

	// If this entity wasn't being tracked in the last update, reset it's 
	// observed signature limits
	if (entityTag.lastObservationTime !== this.$private.lastObservationTime)
	{
		entityTag.minimumObservedSignature = Infinity;
		entityTag.maximumObservedSignature = 0;
	}

	entityTag.minimumObservedSignature = Math.min(entityTag.minimumObservedSignature, signature);
	entityTag.maximumObservedSignature = Math.max(entityTag.maximumObservedSignature, signature);
	entityTag.lastObservationTime = clock.adjustedSeconds;

	// The signature we will use is the average of the observed range
	signature = (entityTag.minimumObservedSignature + entityTag.maximumObservedSignature) / 2;
		
	bogey.signature = signature;
	bogey.expiry = clock.adjustedSeconds + Math.sqrt(this.$monitor._getFieldDisturbanceRange(signature) * 2.0);

	var entityTag = this.$monitor._getEntityTag(entity);
	if (entityTag.timeToMassLock > entityTag.torusFieldInfluence) return bogey;

	// Never fail to ID stars and planets
	if (entity.isSun || entity.isPlanet)
	{
		bogey.identity = "Gravity well";
		return bogey;
	}

	// Binary search of sorted (min to max) array $shipSignatureLibrary for closest 
	// matching signature value
	var bounds = [0, this.$shipSignatureLibrary.length - 1];

	while (bounds[1] - bounds[0] > 1)
	{
		var pivot = Math.floor((bounds[0] + bounds[1]) / 2);
		
		var flagDirection = (this.$shipSignatureLibrary[pivot].signature < signature) ? 0 : 1;

		bounds[flagDirection] = pivot;
	}

	// no ID was made (only if library empty) */
	if (bounds[0] > bounds[1]) return bogey;
	
	var delta =
	[
		Math.abs(this.$shipSignatureLibrary[bounds[0]].signature - signature),
		Math.abs(this.$shipSignatureLibrary[bounds[1]].signature - signature)
	];

	var match = (delta[0] < delta[1])
		? this.$shipSignatureLibrary[bounds[0]]
		: this.$shipSignatureLibrary[bounds[1]]
		;

	var discrepancy = Math.abs(match.signature - bogey.signature) / match.signature;

	

	if (discrepancy < this.$const.maximumDiscrepancy)
	{
		bogey.identity = match.identities[0];
		bogey.discrepancy = discrepancy;
	}

	return bogey;
}

this._setMFDFooterText = function(text)
{
	this.$mfdTabularText[9][0].text = text;
}


/**
 * From the supplied array of entities, calls _identifyBogey on each entity
 * and loads the results into $bogies
 */
this._identifyBogies = function(entities)
{
	var bogies = [];

	//log(this.name, "this._identifyBogies");
	
	for (var i = 0; i < entities.length; ++i)
	{
		var identification;
		
		// only entities with a name can be identified
		if (!this._getNameForEntity(entities[i])) continue;

		// indentification must succeed
		if (!(identification = this._identifyEntity(entities[i]))) continue;

		bogies.push(identification);
	}

	this.$bogies = bogies;

	this.$private.lastObservationTime = clock.adjustedSeconds;
}


this._updateMFD = function(entities)
{
	
	//if (!entities) return;
	//if (entities) this._identifyBogies(entities);
	//$timeToMassLock = this.$monitor.$public.timeToMassLock;
	return;
}


this.shipLaunchedFromStation = function(station)
{
	this._resetMFD();
	this.$torusEverEngaged = false;
}


this.playerWillSaveGame = function()
{
	missionVariables["torus_field_analyser_signature_library"] =
		JSON.stringify(this.$shipSignatureLibrary)
		;
}

this.startUp = function()
{
	/* load all friend scripts */
	for (var key in this.$friend)
	{
		var name = this.$friend[key];
		this.$friend[key] = worldScripts[name];
	}


	var events = worldScripts["torus-field-events.raeddi.alaric"].$event;

	events.onPlayerTargetChanged[this.name] = this._onPlayerTargetChanged.bind(this);


	// store torus field monitor script in $monitor so we call call its functions
	if (!(this.$monitor = worldScripts["torus-field-monitor.raeddi.alaric"]))
	{
		log(this.name, "error: World Script 'torus-field-monitor.raeddi.alaric' not found.");
	}
	
	this.$shipSignatureLibrary = missionVariables["torus_field_analyser_signature_library"] 
		? JSON.parse(missionVariables["torus_field_analyser_signature_library"])
		: []
		;

	/* sort signature array in case it has been fiddled with */
	this.$shipSignatureLibrary.sort(
		function (lhs, rhs)
		{
			return lhs.signature - rhs.signature;
		}
	);

	//log(this.name, JSON.stringify(this.$shipSignatureLibrary));

}


this.startUpComplete = function()
{
	this._resetMFD();
	var text = worldScripts["alaric-oxp-utilities"]._buildTabularText(this.$mfdTabularText);
	player.ship.setMultiFunctionText("torus-field-analyser", text, false);

	//return;

	this.$mfdUpdate = new Timer(
			this,			
			function()
			{
				this._updateBogiesMFD();
				var mfdText = worldScripts["alaric-oxp-utilities"]._buildTabularText(this.$mfdTabularText);
				player.ship.setMultiFunctionText("torus-field-analyser", mfdText, false);
			},
			15.0,	// 15 seconds for intro text
			0.25 // then every pingFrequency seconds
		);	

}