"use strict";
this.name        = "torus-field-marker.raeddi.alaric";
this.author      = "Alaric";
this.copyright   = "2016 Alaric";
this.description = "Disturbance marker system for Torus Field Monitor/Analyser";
this.licence     = "CC BY-NC-SA 3.0 AU";
this.version	 = "1.0";


/**
 * Takes a linear interpolation <factor> (0..1) and returns an interpolation
 * factor based upon Newtonian acceleration up to <balance> followed by
 * Newtonian deceleration. Used to slew the mass-lock waypoint in a visually
 * pleasing way when the mass-lock entity changes.
 */
this._getNewtownianInterpolation = function(factor, balance)
{
	return (factor < balance)
		? Math.pow(factor, 2) / balance
		: (balance + (factor - 2) * factor) / (balance - 1)
		;
}
  
 this.$const =
 {
	nullVector: Vector3D(Infinity, Infinity, Infinity)
 };
 
 this.$private =
 {
	markerEntity:	null
 };
 
 
 /**
  * Sets the primary disturbance vector for display purposes. Note that this 
  * vector is *player relative*. If called with null vector, will clear the
  * primary target.
  */
 this._setPrimaryDisturbance = function(vector)
 {
	var entity;

	if (!(entity = this.$private.markerEntity)) return false;
	
	entity.position = (vector) 
		? player.ship.position.add(vector)
		: this.$const.nullVector
		;
	
	return true;
 }

 this._clearAllDisturbances = function()
 {
	this._setPrimaryDisturbance(null);
 }

  
 this._createPrivateMarkerEntity = function()
 {
	this._deletePrivateMarkerEntity();

	this.$private.markerEntity = system.addVisualEffect(
		"torus-field-marker.raeddi.alaric",
		this.$const.nullVector
		);

	return (!!this.$private.markerEntity);
 }

 this._deletePrivateMarkerEntity = function()
 {
	if (this.$private.markerEntity)
	{
		this.$private.markerEntity.remove(true);
		this.$private.markerEntity = null;
	}
 }


 /* Management of marker creation/destruction. These belong to the current 
  * System, so must be managed correctly for system changes
  */

 this.shipLaunchedFromStation = function(station)
 {
	this._createPrivateMarkerEntity();
 }
 this.shipWillDockWithStation = function(station)
 {
	this._deletePrivateMarkerEntity();
 }

 this.shipWillEnterWitchspace = function(cause)
 {
	this._deletePrivateMarkerEntity();
 }

 this.shipExitedWitchspace = function()
 {
	this._createPrivateMarkerEntity();
 }

 